package com.sillycycle.bagleyd.threed;

/*
 * @(#)ThreedObject.java
 *
 * Copyright 1994 - 2016  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

/**
 * The <code>ThreedObject</code> class handles the object to be
 * manipulated for the <code>ThreedFrame</code> class.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/threed.html
 */

public class ThreedObject {
	String name;
	double scale;
	boolean convex;
	ThreedPoint[] vertices;
	int[] edges;
	ThreedSurface[] surfaces;

	ThreedPoint[] local;
	//ThreedPoint[] world;
	//int[] map;
	ThreedVertex[] vertex;
	double radius;
	//int state;
	ThreedPoint origin = new ThreedPoint(0.0, 0.0, 0.0);
	ThreedAngle orient = new ThreedAngle(0, 0, 0);
	static final boolean DEBUG = true;

	static final double[] SINE =
		createTrigTables(ThreedInterface.RT_ANGLE + 1);

	static final double[] createTrigTables(int n) {
		int angle;
		double[] trig = new double[n];
		double sepRadians = Math.PI / ThreedInterface.ST_ANGLE;

		for (angle = 0; angle <= ThreedInterface.RT_ANGLE; angle++) {
			trig[angle] = Math.sin(sepRadians * angle);
		}
		return trig;
	}

	static double sinInt(int angleDegrees) {
		int angle = (ThreedInterface.NUM_DEGREES + angleDegrees) %
			ThreedInterface.NUM_DEGREES;

		if (angle <= ThreedInterface.RT_ANGLE)
			return SINE[angle];
		else if (angle <= ThreedInterface.ST_ANGLE)
			return SINE[ThreedInterface.ST_ANGLE - angle];
		else if (angle <= ThreedInterface.NUM_DEGREES -
				ThreedInterface.RT_ANGLE)
			return -SINE[angle - ThreedInterface.ST_ANGLE];
		else
			return -SINE[ThreedInterface.NUM_DEGREES - angle];
	}

	static double cosInt(int angleDegrees) {
		int angle = (ThreedInterface.NUM_DEGREES + angleDegrees) %
			ThreedInterface.NUM_DEGREES;

		if (angle <= ThreedInterface.RT_ANGLE)
			return SINE[ThreedInterface.RT_ANGLE - angle];
		else if (angle <= ThreedInterface.ST_ANGLE)
			return -SINE[angle - ThreedInterface.RT_ANGLE];
		else if (angle <= ThreedInterface.NUM_DEGREES -
				ThreedInterface.RT_ANGLE)
			return -SINE[ThreedInterface.NUM_DEGREES -
				ThreedInterface.RT_ANGLE - angle];
		else
			return SINE[angle - ThreedInterface.NUM_DEGREES +
				ThreedInterface.RT_ANGLE];
	}

	public ThreedObject(String name, ThreedPoint origin, double scale, int convex,
			ThreedPoint[] vertices, int[] edges,
			ThreedSurface[] surfaces) {
		this.name = name;
		this.origin = origin;
		this.scale = scale;
		this.convex = (convex != 0);
		this.vertices = vertices;
		this.surfaces = surfaces;
		this.edges = edges;
		//this.surface = new ThreedSurface[surfaces.length];
		this.vertex = new ThreedVertex[vertices.length];
		//this.world = new ThreedPoint[vertices.length];
		this.local = new ThreedPoint[vertices.length];
	}

	/* Black Art of 3D Game Programming Andre LaMothe */

	/* I guess the average radius would be the average of the vertices and
	 * the average of the midpoint of the faces (minimumObjectRadius),
	 * averaged together */

	double minimumObjectRadius(){
		double objectRadius, minRadius, x, y, z;
		int surf, vert, objectVertex;

		minRadius = -1.0;
		for (surf = 0; surf < this.surfaces.length; surf++) {
			int mapIndex = this.surfaces[surf].mapIndex;

			x = y = z = 0.0;
			for (vert = 0; vert < this.surfaces[surf].numVertices;
					vert++, mapIndex++) {
				objectVertex = this.edges[mapIndex];
				x += this.local[objectVertex].x;
				y += this.local[objectVertex].y;
				z += this.local[objectVertex].z;
			}
			if (vert > 0) {
				x /= vert;
				y /= vert;
				z /= vert;
			}
			objectRadius = Math.sqrt(x * x + y * y + z * z);
			if (objectRadius < minRadius || minRadius == -1.0)
				minRadius = objectRadius;
		}
		return minRadius;
	}

	double maximumObjectRadius() {
		double objectRadius, maxRadius, x, y, z;
		int vert;

		maxRadius = 0.0;
		for (vert = 0; vert < this.vertices.length; vert++) {
			x = this.local[vert].x;
			y = this.local[vert].y;
			z = this.local[vert].z;
			objectRadius = Math.sqrt(x * x + y * y + z * z);
			if (objectRadius > maxRadius)
				maxRadius = objectRadius;
		}
		return maxRadius;
	}


	void translateObject(ThreedVector v) {
		this.origin.x += v.x;
		this.origin.y += v.y;
		this.origin.z += v.z;
	}

	void positionObject(ThreedPoint p) {
		this.origin.x = p.x;
		this.origin.y = p.y;
		this.origin.z = p.z;
	}

	void scaleObject(double scaleFactor) {
		int /*plane,*/ point;
		/*double scale;*/ /* holds the square of the scaling factor,
				 needed to resize the surface normal for
				 lighting calculations */

		for (point = 0; point < this.vertices.length; point++) {
			this.local[point].x *= scaleFactor;
			this.local[point].y *= scaleFactor;
			this.local[point].z *= scaleFactor;
		}
		/*scale = scaleFactor * scaleFactor;*/
		/* now scale all pre-computed normals */
		/*for (plane = 0; plane < this.surfaces.length; plane++) {
			this.surfaces[plane].normalLength *= scale;
		}*/
		this.radius *= scaleFactor;
	}

	void rotateObject(ThreedAngle angle) {
		int vert;
		int product = 0; /* used to determine which matrices need
			    multiplying */

		ThreedPoint tempPoint = new ThreedPoint(0.0, 0.0, 0.0);
		ThreedMatrix rotate = new ThreedMatrix();
		ThreedMatrix rotateTheta = new ThreedMatrix();
		ThreedMatrix rotatePhi = new ThreedMatrix();
		ThreedMatrix rotatePsi = new ThreedMatrix();
		ThreedMatrix temp = new ThreedMatrix();

		if (angle.theta == 0 && angle.phi == 0 && angle.psi == 0)
			return;
		orient.theta += angle.theta;
		orient.phi += angle.phi;
		orient.psi += angle.psi;
		if (orient.theta < ThreedInterface.MIN_DEGREES ||
				orient.theta >= ThreedInterface.NUM_DEGREES)
			orient.theta = (ThreedInterface.NUM_DEGREES +
				orient.theta) % ThreedInterface.NUM_DEGREES;
		if (orient.phi < ThreedInterface.MIN_DEGREES ||
				orient.phi >= ThreedInterface.NUM_DEGREES)
			orient.phi = (ThreedInterface.NUM_DEGREES +
				orient.phi) % ThreedInterface.NUM_DEGREES;
		if (orient.psi < ThreedInterface.MIN_DEGREES ||
				orient.psi >= ThreedInterface.NUM_DEGREES)
			orient.psi = (ThreedInterface.NUM_DEGREES +
				orient.psi) % ThreedInterface.NUM_DEGREES;
		rotate.identityMatrix4x4();
		if (angle.phi != 0) {
			rotatePhi.identityMatrix4x4();
			rotatePhi.setInMatrix(1, 1, cosInt(angle.phi));
			rotatePhi.setInMatrix(1, 2, sinInt(angle.phi));
			rotatePhi.setInMatrix(2, 1, -sinInt(angle.phi));
			rotatePhi.setInMatrix(2, 2, cosInt(angle.phi));
			product |= 4;
		}
		if (angle.theta != 0) {
			rotateTheta.identityMatrix4x4();
			rotateTheta.setInMatrix(0, 0, cosInt(angle.theta));
			rotateTheta.setInMatrix(0, 2, -sinInt(angle.theta));
			rotateTheta.setInMatrix(2, 0, sinInt(angle.theta));
			rotateTheta.setInMatrix(2, 2, cosInt(angle.theta));
			product |= 2;
		}
		if (angle.psi != 0) {
			rotatePsi.identityMatrix4x4();
			rotatePsi.setInMatrix(0, 0, cosInt(angle.psi));
			rotatePsi.setInMatrix(0, 1, sinInt(angle.psi));
			rotatePsi.setInMatrix(1, 0, -sinInt(angle.psi));
			rotatePsi.setInMatrix(1, 1, cosInt(angle.psi));
			product |= 1;
		}
		/* Compute the final rotation matrix, determine the proper
		 * product of matrices.  Determine which matrices need
		 * multiplying, this is worth the time it would take to
		 * concatenate matrices together that do not have any
		 * effect. */
		/* System.out.println("product " + product); */
		switch (product) {
		case 0: /* do nothing there is no rotation */
			break;
		case 1: /* final matrix = psi */
			rotate.copyMatrix4x4(rotatePsi);
			break;
		case 2: /* final matrix = theta */
			rotate.copyMatrix4x4(rotateTheta);
			break;
		case 3: /* final matrix = theta * psi */
			rotate.multiplyMatrix4x4_4x4(rotateTheta, rotatePsi);
			break;
		case 4: /* final matrix = phi */
			rotate.copyMatrix4x4(rotatePhi);
			break;
		case 5: /* final matrix = phi * psi */
			rotate.multiplyMatrix4x4_4x4(rotatePhi, rotatePsi);
			break;
		case 6: /* final matrix = phi * theta */
			rotate.multiplyMatrix4x4_4x4(rotatePhi, rotateTheta);
			break;
		case 7: /* final matrix = phi * theta * psi */
			temp.multiplyMatrix4x4_4x4(rotatePhi, rotateTheta);
			rotate.multiplyMatrix4x4_4x4(temp, rotatePsi);
			break;
		default:
			break;
		}
		for (vert = 0; vert < this.vertices.length; vert++) {
			tempPoint.x = this.local[vert].x * rotate.getInMatrix(0, 0) +
				this.local[vert].y * rotate.getInMatrix(1, 0) +
				this.local[vert].z * rotate.getInMatrix(2, 0);
			tempPoint.y = this.local[vert].x * rotate.getInMatrix(0, 1) +
				this.local[vert].y * rotate.getInMatrix(1, 1) +
				this.local[vert].z * rotate.getInMatrix(2, 1);
			tempPoint.z = this.local[vert].x * rotate.getInMatrix(0, 2) +
				this.local[vert].y * rotate.getInMatrix(1, 2) +
				this.local[vert].z * rotate.getInMatrix(2, 2);
			this.local[vert].x = tempPoint.x;
			this.local[vert].y = tempPoint.y;
			this.local[vert].z = tempPoint.z;
			/* System.out.println(vert + ": x " +
				this.local[vert].x + ", y " +
				this.local[vert].y + ", z " +
				this.local[vert].z); */
		}
	}
}
