package com.sillycycle.bagleyd.threed;

/*
 * @(#)ThreedUniverse.java
 *
 * Copyright 2016  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.io.File;
import java.io.IOException;
import java.net.URL;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.ParserConfigurationException;
import org.w3c.dom.DOMException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.w3c.dom.Node;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

/**
 * The <code>ThreedUniverse</code> class handles the objects to be
 * manipulated for the <code>ThreedFrame</code> class.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/threed.html
 */

public class ThreedUniverse {
	ThreedPoint distance;
	ThreedAngle angle;
	ThreedObject[] objects;
	static final boolean DEBUG = false;

	public ThreedUniverse(ThreedPoint distance, ThreedAngle angle,
			ThreedObject[] objects) {
		this.distance = distance;
		this.angle = angle;
		this.objects = objects;
	}

	static ThreedUniverse readObjectsXML(String fileName, URL url) {
		try {
			DocumentBuilderFactory docBuilderFactory =
				DocumentBuilderFactory.newInstance();
			DocumentBuilder docBuilder =
				docBuilderFactory.newDocumentBuilder();
			Document doc;

			if (url == null) {
				doc = docBuilder.parse(new File(fileName));
			} else {
				doc = docBuilder.parse(new URL(url, fileName).openStream());
				//System.out.println("Loaded xml: " + url + fileName);
			}
			// normalize text representation
			doc.getDocumentElement().normalize();
			if (DEBUG)
				System.out.println("Root element of the doc is " +
					doc.getDocumentElement().getNodeName());

			NodeList header =
				doc.getElementsByTagName("header");
			if (DEBUG)
				System.out.println("Total header length: " + header.getLength());
			ThreedPoint distance = null;
			ThreedAngle angle = null;
			ThreedObject[] objects;
			int numObjects = 0;
			for (int h = 0; h < header.getLength(); h++) {
				Node headerNode = header.item(h);
				if (headerNode.getNodeType() == Node.ELEMENT_NODE) {
					Element headerElement = (Element) headerNode;
					NodeList numberList = headerElement.getElementsByTagName("number");
					Element numberElement = (Element)numberList.item(0);

					NodeList textNumberList = numberElement.getChildNodes();
					String number = textNumberList.item(0).getNodeValue().trim();
					numObjects = Integer.parseInt(number);
					distance = readDistance(headerElement);
					if (DEBUG) {
						if (distance == null)
							System.out.println("distance: null");
						else
							System.out.println("distance: "
									+ distance.x + " " + distance.y + " " + distance.z);
					}
					angle = readAngle(headerElement);
					if (DEBUG) {
						if (angle == null)
							System.out.println("angle: null");
						else
							System.out.println("angle: "
									+ angle.theta + " " + angle.phi + " " + angle.psi);
					}
				}
			}
			if (DEBUG)
				System.out.println("Number of Objects = " + numObjects);
			NodeList objectsList =
				doc.getElementsByTagName("objects");
			objects = new ThreedObject[numObjects];
			for (int os = 0; os < objectsList.getLength(); os++) {
				Node objectsNode = objectsList.item(os);
				if (objectsNode.getNodeType() == Node.ELEMENT_NODE) {
					Element objectsElement = (Element) objectsNode;
					NodeList objectList = objectsElement.getElementsByTagName("object");
					for (int o = 0; o < objectList.getLength(); o ++) {
						objects[o] = readObject(objectList.item(o));
					}
				}
			}
			return new ThreedUniverse(distance, angle, objects);
		} catch (SAXParseException err) {
			System.out.println("** Parsing error" + ", line " +
				err.getLineNumber() + ", uri " +
				err.getSystemId());
			System.out.println(" " + err.getMessage());
		} catch (SAXException e) {
			Exception x = e.getException();
			((x == null) ? e : x).printStackTrace();
		} catch (ParserConfigurationException | IOException | DOMException | NumberFormatException e) {
			e.printStackTrace();
		}
		return null;
	}

	static ThreedObject readObject(Node objectNode) {
		try {
			if (objectNode.getNodeType() != Node.ELEMENT_NODE) {
				return null;
			}
			Element objectElement = (Element) objectNode;

			NodeList itemList = objectElement.getElementsByTagName("name");
			Element itemElement = (Element)itemList.item(0);
			NodeList textItemList = itemElement.getChildNodes();
			String name = (textItemList.item(0)).getNodeValue().trim();
			if (DEBUG)
				System.out.println("name: " + name);

			ThreedPoint origin = readOrigin(objectElement);
			if (DEBUG) {
				if (origin == null)
					System.out.println("origin: null");
				else
					System.out.println("origin: "
							+ origin.x + " " + origin.y + " " + origin.z);
			}

			itemList = objectElement.getElementsByTagName("scale");
			itemElement = (Element)itemList.item(0);
			textItemList = itemElement.getChildNodes();
			double scale = Double.parseDouble((textItemList.item(0)).getNodeValue().trim());
			if (DEBUG)
				System.out.println("scale: " + scale);

			itemList = objectElement.getElementsByTagName("convex");
			itemElement = (Element)itemList.item(0);
			textItemList = itemElement.getChildNodes();
			int convex = Integer.parseInt((textItemList.item(0)).getNodeValue().trim());
			if (DEBUG)
				System.out.println("convex: " + convex);

			itemList = objectElement.getElementsByTagName("vertices");
			itemElement = (Element)itemList.item(0);
			ThreedPoint[] vertices = readVertices(itemElement);

			itemList = objectElement.getElementsByTagName("surfaces");
			itemElement = (Element)itemList.item(0);
			ThreedSurface[] surfaces = readSurfaces(itemElement);

			itemList = objectElement.getElementsByTagName("edges");
			itemElement = (Element)itemList.item(0);
			int[] edges = readEdges(itemElement, surfaces);

			return new ThreedObject(name, origin, scale, convex,
				vertices, edges, surfaces);
		} catch (DOMException | NumberFormatException e) {
			e.printStackTrace();
			return null;
		}
	}

	static ThreedPoint readDistance(Element element) {
		try {
			NodeList distanceList = element.getElementsByTagName("distance");
			Element distanceElement = (Element)distanceList.item(0);
			if (distanceElement.getNodeType() != Node.ELEMENT_NODE) {
				return null;
			}
			NodeList itemList;
			Element itemElement;
			NodeList textItemList;

			itemList = distanceElement.getElementsByTagName("distanceX");
			itemElement = (Element)itemList.item(0);

			textItemList = itemElement.getChildNodes();
			double originX = Double.parseDouble((textItemList.item(0)).getNodeValue().trim());

			itemList = distanceElement.getElementsByTagName("distanceY");
			itemElement = (Element)itemList.item(0);
			textItemList = itemElement.getChildNodes();
			double originY = Double.parseDouble((textItemList.item(0)).getNodeValue().trim());

			itemList = distanceElement.getElementsByTagName("distanceZ");
			itemElement = (Element)itemList.item(0);
			textItemList = itemElement.getChildNodes();
			double originZ = Double.parseDouble((textItemList.item(0)).getNodeValue().trim());

			return new ThreedPoint(originX, originY, originZ);
		} catch (DOMException | NumberFormatException e) {
			e.printStackTrace();
			return null;
		}
	}

	static ThreedAngle readAngle(Element element) {
		try {
			NodeList degreesList = element.getElementsByTagName("degrees");
			Element degreesElement = (Element)degreesList.item(0);
			if (degreesElement.getNodeType() != Node.ELEMENT_NODE) {
				return null;
			}
			NodeList itemList;
			Element itemElement;
			NodeList textItemList;

			itemList = degreesElement.getElementsByTagName("degreesTheta");
			itemElement = (Element)itemList.item(0);

			textItemList = itemElement.getChildNodes();
			int theta = Integer.parseInt((textItemList.item(0)).getNodeValue().trim());

			itemList = degreesElement.getElementsByTagName("degreesPhi");
			itemElement = (Element)itemList.item(0);
			textItemList = itemElement.getChildNodes();
			int phi = Integer.parseInt((textItemList.item(0)).getNodeValue().trim());

			itemList = degreesElement.getElementsByTagName("degreesPsi");
			itemElement = (Element)itemList.item(0);
			textItemList = itemElement.getChildNodes();
			int psi = Integer.parseInt((textItemList.item(0)).getNodeValue().trim());

			return new ThreedAngle(theta, phi, psi);
		} catch (DOMException | NumberFormatException e) {
			e.printStackTrace();
			return null;
		}
	}

	static ThreedPoint readOrigin(Element element) {
		try {
			NodeList originList = element.getElementsByTagName("origin");
			Element originElement = (Element)originList.item(0);
			if (originElement.getNodeType() != Node.ELEMENT_NODE) {
				return null;
			}
			NodeList itemList;
			Element itemElement;
			NodeList textItemList;

			itemList = originElement.getElementsByTagName("originX");
			itemElement = (Element)itemList.item(0);

			textItemList = itemElement.getChildNodes();
			double originX = Double.parseDouble((textItemList.item(0)).getNodeValue().trim());

			itemList = originElement.getElementsByTagName("originY");
			itemElement = (Element)itemList.item(0);
			textItemList = itemElement.getChildNodes();
			double originY = Double.parseDouble((textItemList.item(0)).getNodeValue().trim());

			itemList = originElement.getElementsByTagName("originZ");
			itemElement = (Element)itemList.item(0);
			textItemList = itemElement.getChildNodes();
			double originZ = Double.parseDouble((textItemList.item(0)).getNodeValue().trim());

			return new ThreedPoint(originX, originY, originZ);
		} catch (DOMException | NumberFormatException e) {
			e.printStackTrace();
			return null;
		}
	}

	static ThreedPoint[] readVertices(Node verticeNode) {
		try {
			if (verticeNode.getNodeType() != Node.ELEMENT_NODE) {
				return null;
			}
			Element verticeElement = (Element) verticeNode;

			NodeList itemList = verticeElement.getElementsByTagName("number");
			Element itemElement = (Element)itemList.item(0);
			NodeList textItemList = itemElement.getChildNodes();
			int number = Integer.parseInt((textItemList.item(0)).getNodeValue().trim());
			if (DEBUG)
				System.out.println("Vertices: " + number);

			itemList = verticeElement.getElementsByTagName("data");
			itemElement = (Element)itemList.item(0);
			textItemList = itemElement.getChildNodes();

			String xmlDbl = (textItemList.item(0)).getNodeValue();
			String newDbl = xmlDbl.replaceAll("\\r|\\n", " ").replaceAll("\\s+", " ").trim();
			if (DEBUG)
				System.out.println("Data: " + newDbl);
			String[] numbers = newDbl.split(" ");
			int length;
			if (numbers.length == 0 || numbers[0].equals("")) {
				length = 0;
			} else {
				length = numbers.length;
			}
			if (DEBUG)
				System.out.println("Number: " + length);
			if (number * 3 != length) {
				System.out.println("Vertices=" + number
					+ " but data.length= " + length);
				return null;
			}
			ThreedPoint[] threedPoint = new ThreedPoint[number];
			for (int i = 0; i < number; i++) {
				double x, y, z;
				x = Double.parseDouble(numbers[3 * i]);
				y = Double.parseDouble(numbers[3 * i + 1]);
				z = Double.parseDouble(numbers[3 * i + 2]);
				threedPoint[i] = new ThreedPoint(x, y, z);
			}
			return threedPoint;
		} catch (DOMException | NumberFormatException e) {
			e.printStackTrace();
			return null;
		}
	}

	static int[] readEdges(Node edgeNode, ThreedSurface[] surfaces) {
		try {
			if (edgeNode.getNodeType() != Node.ELEMENT_NODE) {
				return null;
			}
			Element edgeElement = (Element) edgeNode;

			NodeList itemList = edgeElement.getElementsByTagName("number");
			Element itemElement = (Element)itemList.item(0);
			NodeList textItemList = itemElement.getChildNodes();
			int number = Integer.parseInt((textItemList.item(0)).getNodeValue().trim());
			if (DEBUG)
				System.out.println("Edges: " + number);

			itemList = edgeElement.getElementsByTagName("data");
			itemElement = (Element)itemList.item(0);
			textItemList = itemElement.getChildNodes();

			String xmlInt = (textItemList.item(0)).getNodeValue();
			String newInt = xmlInt.replaceAll("\\r|\\n", " ").replaceAll("\\s+", " ").trim();
			if (DEBUG)
				System.out.println("Data: " + newInt);
			String[] numbers = newInt.split(" ");
			int length;
			if (numbers.length == 0 || numbers[0].equals("")) {
				length = 0;
			} else {
				length = numbers.length;
			}
			if (DEBUG)
				System.out.println("Number: " + length);
			if (number + surfaces.length  != length) {
				System.out.println("Edges=" + number
					+ " and Surfaces=" + surfaces.length
					+ " but data.length= " + length);
			}

			int index = 0;
			int edge = 0;
			int[] edges = new int[number];
			for (ThreedSurface surface : surfaces) {
				surface.numVertices = Integer.parseInt(numbers[index++]);
				for (int j = 0; j < surface.numVertices; j++) {
					edges[edge++] =
						Integer.parseInt(numbers[index++]);
				}
			}
			return edges;
			/*ArrayList<int[]> arrayListArray = new ArrayList<int[]>();
			int countOuter = -1;
			int countInner = 0;
			int innerLimit = -1;
			int[] array = null;
			for (int i = 0; i < number; i++) {
				int value = Integer.parseInt(numbers[i]);
				if (countInner >= innerLimit) {
					if (value <= 0) {
						System.out.println("size=" + value);
						return null;
					}
					countOuter++;
					countInner = 0;
					innerLimit = value;
					array = new int[innerLimit];
					arrayListArray.add(countOuter, array);
				} else {
					array[countInner++] = value;
				}
			}
			return arrayListArray;*/
		} catch (DOMException | NumberFormatException e) {
			e.printStackTrace();
			return null;
		}
	}

	static ThreedSurface[] readSurfaces(Node surfaceNode) {
		try {
			if (surfaceNode.getNodeType() != Node.ELEMENT_NODE) {
				return null;
			}
			Element surfaceElement = (Element) surfaceNode;

			NodeList itemList = surfaceElement.getElementsByTagName("number");
			Element itemElement = (Element)itemList.item(0);
			NodeList textItemList = itemElement.getChildNodes();
			int number = Integer.parseInt((textItemList.item(0)).getNodeValue().trim());
			if (DEBUG)
				System.out.println("Surfaces: " + number);

			itemList = surfaceElement.getElementsByTagName("data");
			itemElement = (Element)itemList.item(0);
			textItemList = itemElement.getChildNodes();

			String xmlInt = (textItemList.item(0)).getNodeValue();
			String newInt = xmlInt.replaceAll("\\r|\\n", " ").replaceAll("\\s+", " ").trim();
			if (DEBUG)
				System.out.println("Data: " + newInt);
			String[] numbers = newInt.split(" ");
			int length;
			if (numbers.length == 0 || numbers[0].equals("")) {
				length = 0;
			} else {
				length = numbers.length;
			}
			if (number * 3 != length) {
				System.out.println("Surface=" + number
					+ " but data.length= " + length);
				return null;
			}
			if (DEBUG)
				System.out.println("Number: " + length);
			ThreedSurface[]  surfaces = new ThreedSurface[number];
			for (int i = 0; i < number; i++) {
				int mapIndex = Integer.parseInt(numbers[3 * i]);
				int depthIndex = Integer.parseInt(numbers[3 * i + 1]);
				int color = Integer.parseInt(numbers[3 * i + 2]);
				surfaces[i] = new ThreedSurface(mapIndex, depthIndex, color);
			}
			return surfaces;
		} catch (DOMException | NumberFormatException e) {
			e.printStackTrace();
			return null;
		}
	}
}
